/*
  Copyright 2020 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.auth

import android.content.Context
import com.kakao.sdk.auth.model.Prompt
import com.kakao.sdk.auth.model.OAuthToken

/**
 * 카카오 로그인을 실행하기 위한 클라이언트.
 * 카카오톡 또는 기본 웹 브라우저(CustomTabs)를 이용해 카카오계정 정보를 획득하고 API 요청에 필요한 사용자 토큰을 발급 받습니다.
 */
class LoginClient(
    /** @suppress */ val authApiClient: AuthApiClient = AuthApiClient.instance,
    /** @suppress */ val authCodeClient: AuthCodeClient = AuthCodeClient.instance
) {

    /**
     * 카카오톡으로 로그인 가능(설치) 여부 검사.
     */
    fun isKakaoTalkLoginAvailable(context: Context): Boolean =
        authCodeClient.isKakaoTalkLoginAvailable(context)

    /**
     * 카카오톡으로 로그인. 카카오톡에 연결된 카카오계정으로 사용자를 인증하고 [OAuthToken] 발급.
     *
     * 발급된 토큰은 [TokenManagerProvider]에 지정된 토큰 저장소에 자동으로 저장됨.
     *
     * @param context 카카오톡 로그인 Activity를 실행하기 위한 현재 Activity context
     * @param callback 발급 받은 [OAuthToken] 반환.
     */
    @JvmOverloads
    fun loginWithKakaoTalk(
        context: Context,
        requestCode: Int = AuthCodeClient.DEFAULT_REQUEST_CODE,
        channelPublicIds: List<String>? = null,
        serviceTerms: List<String>? = null,
        callback: (token: OAuthToken?, error: Throwable?) -> Unit
    ) {
        val codeVerifier = AuthCodeClient.codeVerifier()
        authCodeClient.authorizeWithKakaoTalk(
            context,
            requestCode,
            channelPublicIds = channelPublicIds,
            serviceTerms = serviceTerms,
            codeVerifier = codeVerifier
        ) { code, codeError ->
            if (codeError != null) {
                callback(null, codeError)
            } else {
                authApiClient.issueAccessToken(code!!, codeVerifier) { token, tokenError ->
                    callback(token, tokenError)
                }
            }
        }
    }

    /**
     * 카카오계정으로 로그인. 기본 웹 브라우저(CustomTabs)에 있는 카카오계정 cookie 로 사용자를 인증하고 [OAuthToken] 발급.
     *
     * 발급된 토큰은 [TokenManagerProvider]에 지정된 토큰 저장소에 자동으로 저장됨.
     *
     * @param context CustomTabs를 실행하기 위한 현재 Activity context
     * @param prompts 동의 화면 요청 시 추가 상호작용을 요청하고자 할 때 전달. [Prompt]
     * @param callback 발급 받은 [OAuthToken] 반환.
     */
    @JvmOverloads
    fun loginWithKakaoAccount(
        context: Context,
        prompts: List<Prompt>? = null,
        channelPublicIds: List<String>? = null,
        serviceTerms: List<String>? = null,
        callback: (token: OAuthToken?, error: Throwable?) -> Unit
    ) {
        val codeVerifier = AuthCodeClient.codeVerifier()
        authCodeClient.authorizeWithKakaoAccount(
            context,
            prompts = prompts,
            channelPublicIds = channelPublicIds,
            serviceTerms = serviceTerms,
            codeVerifier = codeVerifier
        ) { code, codeError ->
            if (codeError != null) {
                callback(null, codeError)
            } else {
                authApiClient.issueAccessToken(code!!, codeVerifier) { token, tokenError ->
                    callback(token, tokenError)
                }
            }
        }
    }

    /**
     * 사용자가 아직 동의하지 않은 개인정보 및 접근권한 동의 항목에 대하여 동의를 요청하는 동의 화면을 출력하고, 사용자 동의 시 동의항목이 업데이트 된 [OAuthToken] 발급.
     *
     * 발급된 토큰은 [TokenManagerProvider]에 지정된 토큰 저장소에 자동으로 저장됨.
     *
     * @param context CustomTabs를 실행하기 위한 현재 Activity context
     * @param scopes 추가로 동의 받고자 하는 동의 항목 ID 목록. 카카오 디벨로퍼스 동의 항목 설정 화면에서 확인 가능.
     * @param callback 발급 받은 [OAuthToken] 반환.s
     */
    fun loginWithNewScopes(
        context: Context,
        scopes: List<String>,
        callback: (token: OAuthToken?, error: Throwable?) -> Unit
    ) {
        authApiClient.agt { agt, agtError ->
            if (agtError != null) {
                callback(null, agtError)
            } else {
                val codeVerifier = AuthCodeClient.codeVerifier()
                authCodeClient.authorizeWithKakaoAccount(
                    context,
                    scopes = scopes,
                    agt = agt,
                    codeVerifier = codeVerifier
                ) { code, codeError ->
                    if (codeError != null) {
                        callback(null, codeError)
                    } else {
                        authApiClient.issueAccessToken(code!!, codeVerifier) { token, tokenError ->
                            callback(token, tokenError)
                        }
                    }
                }
            }
        }
    }

    companion object {
        /**
         * 간편한 API 호출을 위해 기본 제공되는 singleton 객체
         */
        @JvmStatic
        val instance by lazy { LoginClient() }
    }
}